@@ -55,6 +55,10 @@
 #define XIGetKnownProperty(prop) 0
 #endif
 
+#ifndef XI86_SERVER_FD
+#define XI86_SERVER_FD 0x20
+#endif
+
 static int wcmDevOpen(DeviceIntPtr pWcm);
 static int wcmReady(InputInfoPtr pInfo);
 static void wcmDevReadInput(InputInfoPtr pInfo);
@@ -85,29 +89,13 @@ static void wcmKbdCtrlCallback(DeviceIntPtr di, KeybdCtrl* ctrl)
 {
 }
 
-/*****************************************************************************
- * wcmVirtualTabletPadding(InputInfoPtr pInfo)
- ****************************************************************************/
-
-void wcmVirtualTabletPadding(InputInfoPtr pInfo)
-{
-	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
-
-	priv->leftPadding = 0;
-	priv->topPadding = 0;
-
-	if (!is_absolute(pInfo)) return;
-
-	DBG(10, priv, "x=%d y=%d \n", priv->leftPadding, priv->topPadding);
-	return;
-}
-
 /*****************************************************************************
  * wcmInitialToolSize --
  *    Initialize logical size and resolution for individual tool.
  ****************************************************************************/
 
-static void wcmInitialToolSize(InputInfoPtr pInfo)
+TEST_NON_STATIC void
+wcmInitialToolSize(InputInfoPtr pInfo)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
@@ -123,12 +111,18 @@ static void wcmInitialToolSize(InputInfoPtr pInfo)
 	}
 	else
 	{
+		priv->minX = common->wcmMinX;
+		priv->minY = common->wcmMinY;
 		priv->maxX = common->wcmMaxX;
 		priv->maxY = common->wcmMaxY;
 		priv->resolX = common->wcmResolX;
 		priv->resolY = common->wcmResolY;
 	}
 
+	if (!priv->topX)
+		priv->topX = priv->minX;
+	if (!priv->topY)
+		priv->topY = priv->minY;
 	if (!priv->bottomX)
 		priv->bottomX = priv->maxX;
 	if (!priv->bottomY)
@@ -137,18 +131,40 @@ static void wcmInitialToolSize(InputInfoPtr pInfo)
 	return;
 }
 
-static int
-wcmInitAxes(DeviceIntPtr pWcm)
+static void wcmInitAxis(DeviceIntPtr dev, int axis, Atom label, int min, int max, int res, int min_res, int max_res, int mode) {
+	InitValuatorAxisStruct(dev, axis,
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
+	                       label,
+#endif
+	                       min, max, res, min_res, max_res
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
+	                       ,mode
+#endif
+	);
+}
+
+/**
+ * Initialize the device axes with their proper attributes.
+ *
+ * For each axis on the device, we need to provide X with its attributes
+ * so that its values can be interpreted properly. To support older X
+ * servers without axis labels, each axis index has a de-facto meaning.
+ * Any de-facto defined axis index left unused is initialized with default
+ * attributes.
+ */
+static int wcmInitAxes(DeviceIntPtr pWcm)
 {
 	InputInfoPtr pInfo = (InputInfoPtr)pWcm->public.devicePrivate;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 
 	Atom label;
+	int index;
 	int min, max, min_res, max_res, res;
 	int mode;
 
 	/* first valuator: x */
+	index = 0;
 	label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_X);
 	min = priv->topX;
 	max = priv->bottomX;
@@ -157,17 +173,11 @@ wcmInitAxes(DeviceIntPtr pWcm)
 	res = priv->resolX;
 	mode = Absolute;
 
-	InitValuatorAxisStruct(pInfo->dev, 0,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
+
 
 	/* second valuator: y */
+	index = 1;
 	label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_Y);
 	min = priv->topY;
 	max = priv->bottomY;
@@ -176,151 +186,123 @@ wcmInitAxes(DeviceIntPtr pWcm)
 	res = priv->resolY;
 	mode = Absolute;
 
-	InitValuatorAxisStruct(pInfo->dev, 1,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
 
 
 	/* third valuator: pressure */
-
+	index = 2;
+	label = None;
 	mode = Absolute;
 	min_res = max_res = res = 1;
 	min = 0;
+	max = 1;
 
 	if (!IsPad(priv))
 	{
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_PRESSURE);
-		/* pressure normalized to FILTER_PRESSURE_RES */
 		max = FILTER_PRESSURE_RES;
-	} else {
-		/* The pad doesn't have a pressure axis, so initialise third
-		 * axis as unknown absolute axis on the pad. This way, we
-		 * can leave the strip/abswheel axes on later axes and don't
-		 * run the danger of clients misinterpreting the axis info
-		 */
-		label = None;
-		max = 1;
 	}
 
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
 
-	InitValuatorAxisStruct(pInfo->dev, 2,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
 
 	/* fourth valuator: tilt-x, cursor:z-rotation, pad:strip-x */
+	index = 3;
+	label = None;
+	mode = Absolute;
+	min_res = max_res = res = 1;
+	min = 0;
+	max = 1;
 
-	if (IsCursor(priv))
+	if (IsPen(priv))
+	{
+		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_TILT_X),
+		min_res = max_res = res = round(TILT_RES);
+		min = TILT_MIN;
+		max = TILT_MAX;
+	}
+	else if (IsCursor(priv))
 	{
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_RZ);
 		min = MIN_ROTATION;
 		max = MIN_ROTATION + MAX_ROTATION_RANGE - 1;
-		min_res = max_res = res = 1;
-		mode = Absolute;
-	} else if (IsPad(priv))
-	{
-		label = None; /* XXX: what is this axis? */
-		min = 0;
-		max = 1; /* dummy value if !HasFeature(WCM_STRIP) */
-		min_res = max_res = res = 1;
-		mode = Absolute;
-		if (TabletHasFeature(common, WCM_STRIP))
-			max = common->wcmMaxStripX;
-	} else
-	{
-			label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_TILT_X),
-			min = -64;
-			max = 63;
-			min_res = max_res = res = 1;
-			mode = Absolute;
+	}
+	else if (IsPad(priv) && TabletHasFeature(common, WCM_STRIP))
+	{ /* XXX: what is this axis label? */
+		max = common->wcmMaxStripX;
 	}
 
-	InitValuatorAxisStruct(pInfo->dev, 3,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
+
 
 	/* fifth valuator: tilt-y, cursor:throttle, pad:strip-y */
+	index = 4;
+	label = None;
+	mode = Absolute;
+	min_res = max_res = res = 1;
+	min = 0;
+	max = 1;
 
-	if (IsCursor(priv))
+	if (IsPen(priv))
+	{
+		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_TILT_Y);
+		min_res = max_res = res = round(TILT_RES);
+		min = TILT_MIN;
+		max = TILT_MAX;
+	}
+	else if (IsCursor(priv))
 	{
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_THROTTLE);
 		min = -1023;
 		max = 1023;
-		min_res = max_res = res = 1;
-		mode = Absolute;
-	} else if (IsPad(priv))
-	{
-		label = None; /* XXX: what is this axis? */
-		min = 0;
-		max = 1; /* dummy value if !HasFeature(WCM_STRIP) */
-		min_res = max_res = res = 1;
-		mode = Absolute;
-		if (TabletHasFeature(common, WCM_STRIP))
-			max = common->wcmMaxStripY;
-	} else
-	{
-		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_TILT_Y);
-		min = -64;
-		max = 63;
-		min_res = max_res = res = 1;
-		mode = Absolute;
+	}
+	else if (IsPad(priv) && TabletHasFeature(common, WCM_STRIP))
+	{ /* XXX: what is this axis label? */
+		max = common->wcmMaxStripY;
 	}
 
-	InitValuatorAxisStruct(pInfo->dev, 4,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
+
 
 	/* sixth valuator: airbrush: abs-wheel, artpen: rotation, pad:abs-wheel */
+	index = 5;
+	label = None;
+	mode = Absolute;
+	min_res = max_res = res = 1;
+	min = 0;
+	max = 1;
 
 	if (IsStylus(priv))
 	{
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_WHEEL);
 		max = MAX_ROTATION_RANGE + MIN_ROTATION - 1;
 		min = MIN_ROTATION;
-		min_res = max_res = res = 1;
-		mode = Absolute;
-	} else if ((TabletHasFeature(common, WCM_RING)) && IsPad(priv))
+	}
+	else if ((TabletHasFeature(common, WCM_RING)) && IsPad(priv))
 	{
 		/* Touch ring */
 		label = XIGetKnownProperty(AXIS_LABEL_PROP_ABS_WHEEL);
 		min = MIN_PAD_RING;
 		max = MAX_PAD_RING;
-		min_res = max_res = res = 1;
-		mode = Absolute;
 	}
 
-	InitValuatorAxisStruct(pInfo->dev, 5,
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 7
-			       label,
-#endif
-			       min, max, res, min_res, max_res
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 12
-			       , mode
-#endif
-			       );
+	wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
+
+
+	/* seventh valuator: abswheel2 */
+	if ((TabletHasFeature(common, WCM_DUALRING)) && IsPad(priv))
+	{
+		/* XXX: what is this axis label? */
+		index = 6;
+		label = None;
+		mode = Absolute;
+		min_res = max_res = res = 1;
+
+		min = MIN_PAD_RING;
+		max = MAX_PAD_RING;
+
+		wcmInitAxis(pInfo->dev, index, label, min, max, res, min_res, max_res, mode);
+	}
 
 	return TRUE;
 }
@@ -334,6 +316,7 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 {
 	InputInfoPtr pInfo = (InputInfoPtr)pWcm->public.devicePrivate;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
+	WacomCommonPtr common =	priv->common;
 	unsigned char butmap[WCM_MAX_BUTTONS+1];
 	int nbaxes, nbbuttons, nbkeys;
 	int loop;
@@ -349,14 +332,12 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 	nbaxes = priv->naxes;       /* X, Y, Pressure, Tilt-X, Tilt-Y, Wheel */
 	nbbuttons = priv->nbuttons; /* Use actual number of buttons, if possible */
 
+	if (IsPad(priv) && TabletHasFeature(priv->common, WCM_DUALRING))
+		nbaxes = priv->naxes = nbaxes + 1; /* ABS wheel 2 */
+
 	/* if more than 3 buttons, offset by the four scroll buttons,
 	 * otherwise, alloc 7 buttons for scroll wheel. */
-	nbbuttons = (nbbuttons > 3) ? nbbuttons + 4 : 7;
-
-	/* make sure nbbuttons stays in the range */
-	if (nbbuttons > WCM_MAX_BUTTONS)
-		nbbuttons = WCM_MAX_BUTTONS;
-
+	nbbuttons = min(max(nbbuttons + 4, 7), WCM_MAX_BUTTONS);
 	nbkeys = nbbuttons;         /* Same number of keys since any button may be 
 	                             * configured as an either mouse button or key */
 
@@ -404,8 +385,8 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 			return FALSE;
 	}
 
-	if (!nbaxes || nbaxes > 6)
-		nbaxes = priv->naxes = 6;
+	if (!nbaxes || nbaxes > 7)
+		nbaxes = priv->naxes = 7;
 
 	/* axis_labels is just zeros, we set up each valuator with the
 	 * correct property later */
@@ -432,10 +413,22 @@ static int wcmDevInit(DeviceIntPtr pWcm)
 		return FALSE;
 	}
 
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 16
+	if (IsTouch(priv)) {
+		if (!InitTouchClassDeviceStruct(pInfo->dev, common->wcmMaxContacts,
+						TabletHasFeature(common, WCM_LCD) ? XIDirectTouch : XIDependentTouch,
+						2))
+		{
+			xf86Msg(X_ERROR, "Unable to init touch class device struct!\n");
+			return FALSE;
+		}
+		priv->common->touch_mask = valuator_mask_new(2);
+	}
+#endif
+
 	if (!IsPad(priv))
 	{
 		wcmInitialToolSize(pInfo);
-		wcmMappingFactor(pInfo);
 	}
 
 	if (!wcmInitAxes(pWcm))
@@ -485,7 +478,7 @@ Bool wcmIsWacomDevice (char* fname)
 char *wcmEventAutoDevProbe (InputInfoPtr pInfo)
 {
 	/* We are trying to find the right eventX device */
-	int i, wait = 0;
+	int i = 0, wait = 0;
 	const int max_wait = 2000;
 
 	/* If device is not available after Resume, wait some ms */
@@ -505,7 +498,7 @@ char *wcmEventAutoDevProbe (InputInfoPtr pInfo)
 				xf86ReplaceStrOption(pInfo->options, "Device", fname);
 
 				/* this assumes there is only one Wacom device on the system */
-				return xf86FindOptionValue(pInfo->options, "Device");
+				return xf86CheckStrOption(pInfo->options, "Device", NULL);
 			}
 		}
 		wait += 100;
@@ -515,7 +508,7 @@ char *wcmEventAutoDevProbe (InputInfoPtr pInfo)
 	xf86Msg(X_ERROR, "%s: no Wacom event device found (checked %d nodes, waited %d msec)\n",
 		pInfo->name, i + 1, wait);
 	xf86Msg(X_ERROR, "%s: unable to probe device\n", pInfo->name);
-	return FALSE;
+	return NULL;
 }
 
 /*****************************************************************************
@@ -540,6 +533,22 @@ Bool wcmOpen(InputInfoPtr pInfo)
 	return Success;
 }
 
+/*****************************************************************************
+ * wcmClose --
+ ****************************************************************************/
+
+void wcmClose(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
+
+	DBG(1, priv, "closing device file\n");
+
+	if (pInfo->fd > -1 && !(pInfo->flags & XI86_SERVER_FD)) {
+		xf86CloseSerial(pInfo->fd);
+		pInfo->fd = -1;
+	}
+}
+
 /*****************************************************************************
  * wcmDevOpen --
  *    Open the physical device and init information structs.
@@ -555,18 +564,17 @@ static int wcmDevOpen(DeviceIntPtr pWcm)
 
 	DBG(10, priv, "\n");
 
+	/* If fd management is done by the server, skip common fd handling */
+	if (pInfo->flags & XI86_SERVER_FD)
+		goto got_fd;
+
 	/* open file, if not already open */
 	if (common->fd_refs == 0)
 	{
 		if ((wcmOpen (pInfo) != Success) || !common->device_path)
 		{
 			DBG(1, priv, "Failed to open device (fd=%d)\n", pInfo->fd);
-			if (pInfo->fd >= 0)
-			{
-				DBG(1, priv, "Closing device\n");
-				xf86CloseSerial(pInfo->fd);
-			}
-			pInfo->fd = -1;
+			wcmClose(pInfo);
 			return FALSE;
 		}
 
@@ -592,6 +600,7 @@ static int wcmDevOpen(DeviceIntPtr pWcm)
 		common->fd_refs++;
 	}
 
+got_fd:
 	/* start the tablet data */
 	if (model->Start && (model->Start(pInfo) != Success))
 		return !Success;
@@ -667,7 +676,8 @@ void wcmReadPacket(InputInfoPtr pInfo)
 		/* for all other errors, hope that the hotplugging code will
 		 * remove the device */
 		if (errno != EAGAIN && errno != EINTR)
-			xf86Msg(X_ERROR, "%s: Error reading wacom device : %s\n", pInfo->name, strerror(errno));
+			LogMessageVerbSigSafe(X_ERROR, 0,
+					      "%s: Error reading wacom device : %s\n", pInfo->name, strerror(errno));
 		return;
 	}
 
@@ -735,16 +745,17 @@ static void wcmDevClose(InputInfoPtr pInfo)
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 
+	/* If fd management is done by the server, skip common fd handling */
+	if (pInfo->flags & XI86_SERVER_FD)
+		return;
+
 	DBG(4, priv, "Wacom number of open devices = %d\n", common->fd_refs);
 
 	if (pInfo->fd >= 0)
 	{
 		pInfo->fd = -1;
 		if (!--common->fd_refs)
-		{
-			DBG(1, common, "Closing device; uninitializing.\n");
-			xf86CloseSerial (common->fd);
-		}
+			wcmClose(pInfo);
 	}
 }
 
@@ -859,7 +870,10 @@ static int wcmDevProc(DeviceIntPtr pWcm, int what)
 			}
 			pWcm->public.on = FALSE;
 			break;
-
+#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) * 100 + GET_ABI_MINOR(ABI_XINPUT_VERSION) >= 1901
+		case DEVICE_ABORT:
+			break;
+#endif
 		default:
 			xf86Msg(X_ERROR, "%s: invalid mode=%d. This is an X server bug.\n",
 				pInfo->name, what);
