@@ -17,7 +17,6 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#define WACOM_TOOLS
 #include "config.h"
 #endif
 
@@ -44,6 +43,9 @@
 #define TRACE(...) \
 	if (verbose) fprintf(stderr, "... " __VA_ARGS__)
 
+#define safe_realloc(p, n, s) \
+	((size_t)-1 / (n) < (s) ? NULL : realloc((p), (n)*(s)))
+
 static int verbose = False;
 
 enum printformat {
@@ -203,6 +205,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
+	{
+		.name = "HWTouchSwitchState",
+		.desc = "Touch events turned on/off by hardware switch. ",
+		.prop_name = WACOM_PROP_HARDWARE_TOUCH,
+		.prop_format = 8,
+		.prop_offset = 0,
+		.arg_count = 1,
+		.prop_flags = PROP_FLAG_READONLY | PROP_FLAG_BOOLEAN
+	},
 	{
 		.name = "Gesture",
 		.desc = "Turns on/off multi-touch gesture events "
@@ -432,6 +443,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_READONLY
 	},
+	{
+		.name = "PressureRecalibration",
+		.desc = "Turns on/off Tablet pressure recalibration",
+		.prop_name = WACOM_PROP_PRESSURE_RECAL,
+		.prop_format = 8,
+		.prop_offset = 0,
+		.arg_count = 1,
+		.prop_flags = PROP_FLAG_BOOLEAN
+	},
 	{
 		.name = "MapToOutput",
 		.desc = "Map the device to the given output. ",
@@ -445,13 +465,13 @@ static param_t parameters[] =
 		.get_func = get_all,
 		.prop_flags = PROP_FLAG_READONLY,
 	},
-	{ NULL }
+	{}
 };
 
 /**
  * Deprecated parameters and their respective replacements.
  */
-struct deprecated
+static struct deprecated
 {
 	const char *name;
 	const char *replacement;
@@ -570,6 +590,43 @@ static struct modifier specialkeys[] = {
 
 	{"PgUp", "Prior"}, {"PgDn", "Next"},
 
+	{"del", "Delete"}, {"home", "Home"},
+	{"end", "End"},
+
+	{"`", "quoteleft"},
+	{"-", "minus"},
+	{"=", "equal"},
+	{"[", "bracketleft"},
+	{"]", "bracketright"},
+	{"\\", "backslash"},
+	{";", "semicolon"},
+	{"'", "apostrophe"},
+	{",", "comma"},
+	{".", "period"},
+	{"/", "slash"},
+
+	{"~", "asciitilde"},
+	{"!", "exclam"},
+	{"@", "at"},
+	{"#", "numbersign"},
+	{"$", "dollar"},
+	{"%", "percent"},
+	{"^", "asciicircum"},
+	{"&", "ampersand"},
+	{"*", "asterisk"},
+	{"(", "parenleft"},
+	{")", "parenright"},
+	{"_", "underscore"},
+	{"+", "plus"},
+	{"{", "braceleft"},
+	{"}", "braceright"},
+	{"|", "bar"},
+	{":", "colon"},
+	{"\"", "quotedbl"},
+	{"<", "less"},
+	{">", "greater"},
+	{"?", "question"},
+
 	{ NULL, NULL }
 };
 
@@ -600,6 +657,7 @@ static void print_value(param_t *param, const char *msg, ...)
 			vprintf(msg, va_args);
 			printf("\"\n");
 			break;
+		case FORMAT_DEFAULT:
 		default:
 			vprintf(msg, va_args);
 			printf("\n");
@@ -741,7 +799,7 @@ static void list_one_device(Display *dpy, XDeviceInfo *info)
 			if (atoms[j] == wacom_prop)
 				break;
 
-		if (j <= natoms)
+		if (j < natoms)
 		{
 			unsigned char	*data;
 			Atom		type;
@@ -884,16 +942,16 @@ static int is_modifier(const char* keysym)
 	return (m->name != NULL);
 }
 
-static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data);
-static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data);
-static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
-static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
-static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
+static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data, const size_t size);
+static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data, const size_t size);
+static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
+static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
+static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 
 /* Valid keywords for the --set ButtonX options */
-struct keywords {
+static struct keywords {
 	const char *keyword;
-	int (*func)(Display*, int, char **, unsigned long*, unsigned long *);
+	int (*func)(Display*, int, char **, unsigned long*, unsigned long *, const size_t size);
 } keywords[] = {
 	{"key", special_map_keystrokes},
 	{"button", special_map_button},
@@ -905,7 +963,7 @@ struct keywords {
 
 /* the "core" keyword isn't supported anymore, we just have this here to
    tell people that. */
-static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
 {
 	static int once_only = 1;
 	if (once_only)
@@ -917,8 +975,12 @@ static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *
 	return 0;
 }
 
-static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
 {
+	if (*ndata + 1 > size) {
+		fprintf(stderr, "Insufficient space to store all commands.\n");
+		return 0;
+	}
 	data[*ndata] = AC_MODETOGGLE;
 
 	*ndata += 1;
@@ -926,12 +988,17 @@ static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned
 	return 0;
 }
 
-static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+/* the "displaytoggle" keyword isn't supported anymore, we just have this
+   here to tell people that. */
+static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
 {
-	data[*ndata] = AC_DISPLAYTOGGLE;
-
-	*ndata += 1;
-
+	static int once_only = 1;
+	if (once_only)
+	{
+		printf ("Note: The \"displaytoggle\" keyword is not supported "
+			"anymore and will be ignored.\n");
+		once_only = 0;
+	}
 	return 0;
 }
 
@@ -948,7 +1015,7 @@ static inline int is_valid_keyword(const char *keyword)
 	return 0;
 }
 
-static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
 {
 	int nitems = 0;
 	int i;
@@ -976,13 +1043,22 @@ static int special_map_button(Display *dpy, int argc, char **argv, unsigned long
 			need_press = need_release = 1;
 
 		if (sscanf(btn, "%d", &button) != 1)
-			return nitems;
+			break;
 
 
 		TRACE("Button map %d [%s,%s]\n", abs(button),
 				need_press ?  "press" : "",
 				need_release ?  "release" : "");
 
+		if (need_press && need_release && *ndata + nitems + 2 > size) {
+			fprintf(stderr, "Insufficient space to store all commands.\n");
+			break;
+		}
+		else if ((need_press || need_release) && *ndata + nitems + 1 > size) {
+			fprintf(stderr, "Insufficient space to store all commands.\n");
+			break;
+		}
+
 		if (need_press)
 			data[*ndata + nitems++] = AC_BUTTON | AC_KEYBTNPRESS | abs(button);
 		if (need_release)
@@ -990,7 +1066,7 @@ static int special_map_button(Display *dpy, int argc, char **argv, unsigned long
 	}
 
 	*ndata += nitems;
-	return nitems;
+	return i;
 }
 
 /* Return the first keycode to have the required keysym in the current group.
@@ -1012,23 +1088,22 @@ static int keysym_to_keycode(Display *dpy, KeySym sym)
 
 	for (kc = xkb->min_key_code; kc <= xkb->max_key_code; kc++)
 	{
-		KeySym* ks;
 		int i;
 
-		ks = XkbKeySymsPtr(xkb, kc);
 		for (i = 0; i < XkbKeyGroupWidth(xkb, kc, state.group); i++)
-			if (ks[i] == sym)
+			if (XkbKeycodeToKeysym(dpy, kc, state.group, i) == sym)
 				goto out;
 	}
 
 out:
 	return kc;
 }
+
 /*
    Map gibberish like "ctrl alt f2" into the matching AC_KEY values.
    Returns 1 on success or 0 otherwise.
  */
-static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data)
+static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data, const size_t size)
 {
 	int i;
 	int nitems = 0;
@@ -1076,9 +1151,26 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
 		} else
 			need_press = need_release = 1;
 
+
 		ks = XStringToKeysym(key);
+		if (ks == 0) {
+			key = convert_specialkey(key);
+			ks = XStringToKeysym(key);
+			if (ks == 0)
+				fprintf(stderr, "Warning: unable to map '%s' to a keycode.\n", key);
+		}
+
 		kc = keysym_to_keycode(dpy, ks);
 
+		if (need_press && need_release && *ndata + nitems + 2 > size) {
+			fprintf(stderr, "Insufficient space to store all commands.\n");
+			break;
+		}
+		else if ((need_press || need_release) && *ndata + nitems + 1 > size) {
+			fprintf(stderr, "Insufficient space to store all commands.\n");
+			break;
+		}
+
 		if (need_press)
 			data[*ndata + nitems++] = AC_KEY | AC_KEYBTNPRESS | kc;
 		if (need_release)
@@ -1102,33 +1194,29 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
  */
 static char** strjoinsplit(int argc, char **argv, int *nwords)
 {
-	char buff[1024] = { 0 };
-	char **words	= NULL;
-	char *tmp, *tok;
-
-	while(argc--)
-	{
-		if (strlen(buff) + strlen(*argv) + 1 >= sizeof(buff))
-			break;
-
-		strcat(buff, *argv);
-		strcat(buff, " ");
-		argv++;
-	}
+	char **words = NULL;
+	int i, n;
 
 	*nwords = 0;
+	for (i = 0; i < argc; i++) {
+		char *tok = strtok(argv[i], " ");
+		while (tok) {
+			char **p = safe_realloc(words, *nwords+1, sizeof(char*));
+			if (!p) {
+				fprintf(stderr, "Unable to reallocate memory.\n");
+				return words;
+			}
 
-	for (tmp = buff; tmp && *tmp != '\0'; tmp = index((const char*)tmp, ' ') + 1)
-		(*nwords)++;
+			words = p;
+			words[*nwords] = strdup(tok);
+			if (!words[*nwords]) {
+				fprintf(stderr, "Unable to allocate memory.\n");
+				return words;
+			}
 
-	words = calloc(*nwords, sizeof(char*));
-
-	*nwords = 0;
-	tok = strtok(buff, " ");
-	while(tok)
-	{
-		words[(*nwords)++] = strdup(tok);
-		tok = strtok(NULL, " ");
+			(*nwords)++;
+			tok = strtok(NULL, " ");
+		}
 	}
 
 	return words;
@@ -1149,7 +1237,7 @@ static char** strjoinsplit(int argc, char **argv, int *nwords)
  * @param data  Parsed action data
  * @return 'true' if the whole string was parsed sucessfully, else 'false'
  */
-static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* data, unsigned long *nitems)
+static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* data, unsigned long *nitems, const size_t size)
 {
 	int  i = 0;
 	int  nwords = 0;
@@ -1160,7 +1248,7 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 
 	if (nwords==1 && sscanf(words[0], "%d", &i) == 1)
 	{ /* Mangle "simple" button maps into proper actions */
-		char **new_words = realloc(words, 2);
+		char **new_words = realloc(words, sizeof(char*)*2);
 		if (new_words == NULL)
 		{
 			fprintf(stderr, "Unable to reallocate memory.\n");
@@ -1175,19 +1263,19 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 		nwords = 2;
 	}
 
-	for (i = 0; i < nwords; i++)
+	for (i = 0; i < nwords && *nitems < size; i++)
 	{
 		int j = 0;
 		int keyword_found = 0;
 
-		while (keywords[j].keyword && i < nwords)
+		while (keywords[j].keyword && i < nwords && *nitems < size)
 		{
 			int parsed = 0;
 			if (strcasecmp(words[i], keywords[j].keyword) == 0)
 			{
 				parsed = keywords[j].func(dpy, nwords - i - 1,
 							  &words[i + 1],
-							  nitems, data);
+							  nitems, data, size);
 				i += parsed;
 				keyword_found = 1;
 			}
@@ -1200,6 +1288,7 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 		if (!keyword_found)
 		{
 			fprintf(stderr, "Cannot parse keyword '%s' at position %d\n", words[i], i+1);
+			free(words);
 			return False;
 		}
 	}
@@ -1216,9 +1305,10 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
  * the driver can understand.
  *
  * Once we have a list of actions, we can store it in the appropriate
- * child property. If none exists, we must first create one and update
- * the parent list. If we want no action to occur, we can delete the
- * child property and have the parent point to '0' instead.
+ * child property. Action atoms should be pre-created by the server,
+ * so it is an error if one does not exist. To reset the action to its
+ * default, we can have the parent point to '0' (which is a special
+ * signal to the Wacom driver).
  *
  * @param  dpy         X display we want to query
  * @param  dev         X device we want to modify
@@ -1235,8 +1325,8 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 	unsigned long btnact_nitems, bytes_after;
 	unsigned long nitems = 0;
 
-	data = calloc(256, sizeof(long));
-	if (!parse_actions(dpy, argc, argv, data, &nitems))
+	data = calloc(256, sizeof(unsigned long));
+	if (!parse_actions(dpy, argc, argv, data, &nitems, 256))
 		goto out;
 
 	/* obtain the button actions Atom */
@@ -1244,7 +1334,7 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 				AnyPropertyType, &type, &format, &btnact_nitems,
 				&bytes_after, (unsigned char**)&btnact_data);
 
-	if (offset > btnact_nitems)
+	if (offset >= btnact_nitems)
 	{
 		fprintf(stderr, "Invalid offset into %s property.\n", XGetAtomName(dpy, btnact_prop));
 		goto out;
@@ -1259,16 +1349,20 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 
 	/* set or unset the property */
 	prop = btnact_data[offset];
-	if (nitems > 0)
-	{ /* Setting a new or existing property */
-		if (!prop)
-		{
-			char buff[64];
-			sprintf(buff, "Wacom button action %d", (offset + 1));
-			prop = XInternAtom(dpy, buff, False);
-			btnact_data[offset] = prop;
-		}
+	if (!prop)
+	{
+		/* The subproperty at the given offset is set to 'None',
+		 * meaning the device does not support its meaning. E.g.
+		 * buttons 4-7 are 'None' since the device doesn't have
+		 * physical buttons relating to them.
+		 */
+		fprintf(stderr, "Unsupported offset into '%s' property.\n",
+			XGetAtomName(dpy, btnact_prop));
+		goto out;
+	}
 
+	if (nitems > 0)
+	{ /* Setting an existing property */
 		/* FIXME: the property containing the key sequence must be
 		 * set before updating the button action properties */
 		XChangeDeviceProperty(dpy, dev, prop, XA_INTEGER, 32,
@@ -1288,8 +1382,6 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 					PropModeReplace,
 					(unsigned char*)btnact_data,
 					btnact_nitems);
-
-		XDeleteDeviceProperty(dpy, dev, prop);
 	}
 
 	XFlush(dpy);
@@ -1541,8 +1633,8 @@ static void set(Display *dpy, int argc, char **argv)
 	long *n;
 	char *b;
 	int i;
-	char **values;
-	int nvals;
+	char **values = NULL;
+	int nvals = 0;
 
 	if (argc < 2)
 	{
@@ -1647,10 +1739,10 @@ static void set(Display *dpy, int argc, char **argv)
 				PropModeReplace, data, nitems);
 	XFlush(dpy);
 
+out:
 	for (i = 0; i < nvals; i++)
 		free(values[i]);
 	free(values);
-out:
 	XCloseDevice(dpy, dev);
 	free(data);
 }
@@ -1662,14 +1754,15 @@ static void get_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char
 	XValuatorInfoPtr v;
 
 	info = XListInputDevices(dpy, &ndevices);
-	while(ndevices--)
+
+	for (i = 0; i < ndevices; i++)
 	{
-		d = &info[ndevices];
+		d = &info[i];
 		if (d->id == dev->device_id)
 			break;
 	}
 
-	if (!ndevices) /* device id 0 is reserved and can't be our device */
+	if (i >= ndevices)
 	{
 		fprintf(stderr, "Unable to locate device.\n");
 		return;
@@ -1813,7 +1906,7 @@ static int get_actions(Display *dpy, XDevice *dev,
 				if (last_type != current_type)
 					strcat(buff, "key ");
 				is_press = !!(action & AC_KEYBTNPRESS);
-				detail = XKeycodeToKeysym(dpy, detail, 0);
+				detail = XkbKeycodeToKeysym(dpy, detail, 0, 0);
 				break;
 			case AC_BUTTON:
 				if (last_type != current_type)
@@ -2002,7 +2095,7 @@ static Bool need_xinerama(Display *dpy)
  * @param y_org[out]   Offset from the desktop origin to the mapped area's top edge
  * @return             True if the function could determine the mapped area
  */
-Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x_org, int *y_org)
+static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x_org, int *y_org)
 {
 	Atom matrix_prop = XInternAtom(dpy, "Coordinate Transformation Matrix", True);
 	Atom type;
@@ -2026,7 +2119,7 @@ Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x
 	                   AnyPropertyType, &type, &format, &nitems,
 	                   &bytes_after, (unsigned char**)&data);
 
-	if (format != 32 || type != XInternAtom(dpy, "FLOAT", True))
+	if (format != 32 || type != XInternAtom(dpy, "FLOAT", True) || nitems != 9)
 	{
 		fprintf(stderr,"Property for '%s' has unexpected type - this is a bug.\n",
 			"Coordinate Transformation Matrix");
@@ -2174,14 +2267,14 @@ static Bool set_output_area(Display *dpy, XDevice *dev,
  */
 static Bool set_output_xrandr(Display *dpy, XDevice *dev, char *output_name)
 {
-	int i, found = 0;
+	int i, success = 0;
 	int x, y, width, height;
 	XRRScreenResources *res;
 	XRROutputInfo *output_info;
 	XRRCrtcInfo *crtc_info;
 
 	res = XRRGetScreenResources(dpy, DefaultRootWindow(dpy));
-	for (i = 0; i < res->noutput && !found; i++)
+	for (i = 0; i < res->noutput; i++)
 	{
 		output_info = XRRGetOutputInfo(dpy, res, res->outputs[i]);
 
@@ -2202,24 +2295,18 @@ static Bool set_output_xrandr(Display *dpy, XDevice *dev, char *output_name)
 
 		if (strcmp(output_info->name, output_name) == 0)
 		{
-			found = 1;
+			TRACE("Setting CRTC %s\n", output_name);
+			success = set_output_area(dpy, dev, x, y, width, height);
 			break;
 		}
 	}
 	XRRFreeScreenResources(res);
 
-	/* crtc holds our screen info, need to compare to actual screen size */
-	if (found)
-	{
-		TRACE("Setting CRTC %s\n", output_name);
-		return set_output_area(dpy, dev, x, y, width, height);
-	} else
-	{
+	if (!success)
 		printf("Unable to find output '%s'. "
 			"Output may not be connected.\n", output_name);
 
-		return False;
-	}
+	return success;
 }
 
 /**
@@ -2674,15 +2761,36 @@ static void test_convert_specialkey(void)
 	char buff[5];
 	struct modifier *m;
 
-	/* make sure at least the default keys (ascii 33 - 126) aren't
-	 * specialkeys */
-	for (i = '!'; i <= '~'; i++)
+	/* make sure a-zA-Z aren't specialkeys */
+	for (i = 'a'; i <= 'z'; i++)
 	{
 		sprintf(buff, "%c", i);
 		converted = convert_specialkey(buff);
 		assert(strcmp(converted, buff) == 0);
 	}
 
+	for (i = 'A'; i <= 'Z'; i++)
+	{
+		sprintf(buff, "%c", i);
+		converted = convert_specialkey(buff);
+		assert(strcmp(converted, buff) == 0);
+	}
+
+	/* punctuation are specialkeys */
+	for (i = '!'; i <= '/'; i++)
+	{
+		sprintf(buff, "%c", i);
+		converted = convert_specialkey(buff);
+		assert(strcmp(converted, buff) != 0);
+	}
+	for (i = ':'; i <= '?'; i++)
+	{
+		sprintf(buff, "%c", i);
+		converted = convert_specialkey(buff);
+		assert(strcmp(converted, buff) != 0);
+	}
+
+
 	for (m = specialkeys; m->name; m++)
 	{
 		converted = convert_specialkey(m->name);
@@ -2697,7 +2805,7 @@ static void test_parameter_number(void)
 	 * deprecated them.
 	 * Numbers include trailing NULL entry.
 	 */
-	assert(ARRAY_SIZE(parameters) == 37);
+	assert(ARRAY_SIZE(parameters) == 39);
 	assert(ARRAY_SIZE(deprecated_parameters) == 17);
 }
 
